import re
from io import BytesIO
from pathlib import Path

import discord
import lavalink
import yaml
from red_commons.logging import getLogger

from redbot.core import commands
from redbot.core.data_manager import cog_data_path
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import box, humanize_list, inline

from ..abc import MixinMeta
from ..cog_utils import CompositeMetaClass
from ...managed_node import version_pins
from ...utils import (
    MAX_JAVA_RAM,
    DEFAULT_LAVALINK_YAML,
    DEFAULT_LAVALINK_SETTINGS,
    change_dict_naming_convention,
    has_managed_server,
    has_unmanaged_server,
    sizeof_fmt,
    get_max_allocation_size,
)

log = getLogger("red.cogs.Audio.cog.Commands.lavalink_setup")
_ = Translator("Audio", Path(__file__))


class LavalinkSetupJavaCommand(commands.Command):
    def format_text_for_context(self, ctx: commands.Context, text: str) -> str:
        text = super().format_text_for_context(ctx, text)
        return text.format(
            supported_java_versions=humanize_list(
                list(map(str, version_pins.SUPPORTED_JAVA_VERSIONS))
            ),
        )


class LavalinkSetupCommands(MixinMeta, metaclass=CompositeMetaClass):
    @commands.group(name="llset")
    @commands.is_owner()
    @commands.bot_has_permissions(embed_links=True)
    async def command_llset(self, ctx: commands.Context):
        """`Dangerous commands` Manage Lavalink node configuration settings.

        This command block holds all commands to configure an unmanaged (user maintained) or managed (bot maintained) Lavalink node.

        You should not mess with any command in here unless you have a valid reason to,
        i.e. been told by someone in the Red-Discord Bot support server to do so.
        All the commands in here have the potential to break the Audio cog.
        """

    @command_llset.command(name="java", cls=LavalinkSetupJavaCommand)
    @has_managed_server()
    async def command_llset_java(self, ctx: commands.Context, *, java_path: str = "java"):
        """Change your Java executable path.

        This command shouldn't need to be used most of the time, and is only useful if the host machine has conflicting Java versions.

        If changing this make sure that the Java executable you set is supported by Audio.
        The current supported versions are Java {supported_java_versions}.

        Enter nothing or "java" to reset it back to default.
        """
        if java_path == "java":
            await self.config.java_exc_path.clear()
            await self.send_embed_msg(
                ctx,
                title=_("Java Executable Reset"),
                description=_(
                    "Audio will now use `java` to run your managed Lavalink node. "
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )
        else:
            exc = Path(java_path)
            exc_absolute = exc.absolute()
            if not exc.exists() or not exc.is_file():
                return await self.send_embed_msg(
                    ctx,
                    title=_("Invalid Environment"),
                    description=_("`{java_path}` is not a valid executable").format(
                        java_path=exc_absolute
                    ),
                )
            await self.config.java_exc_path.set(str(exc_absolute))
            await self.send_embed_msg(
                ctx,
                title=_("Java Executable Changed"),
                description=_(
                    "Audio will now use `{exc}` to run your managed Lavalink node. "
                    "Run `{p}{cmd}` for it to take effect."
                ).format(
                    exc=exc_absolute,
                    p=ctx.prefix,
                    cmd=self.command_audioset_restart.qualified_name,
                ),
            )

    @command_llset.command(name="heapsize", aliases=["hs", "ram", "memory"])
    @has_managed_server()
    async def command_llset_heapsize(self, ctx: commands.Context, size: str = MAX_JAVA_RAM):
        """Set the managed Lavalink node maximum heap-size.

        By default, this value is 50% of available RAM in the host machine represented by [1-1024][M|G] (256M, 256G for example)

        This value only represents the maximum amount of RAM allowed to be used at any given point, and does not mean that the managed Lavalink node will always use this amount of RAM.

        To reset this value to the default, run the command without any input.
        """

        async def validate_input(cog, arg):
            match = re.match(r"^(\d+)([MG])$", arg, flags=re.IGNORECASE)
            if not match:
                await ctx.send(_("Heap-size must be a valid measure of size, e.g. 256M, 256G"))
                return 0
            input_in_bytes = int(match.group(1)) * 1024 ** (
                2 if match.group(2).lower() == "m" else 3
            )
            if input_in_bytes < 64 * 1024**2:
                await ctx.send(
                    _(
                        "Heap-size must be at least 64M, however it is recommended to have it set to at least 1G."
                    )
                )
                return 0
            elif (
                input_in_bytes
                > (meta := get_max_allocation_size(cog.managed_node_controller._java_exc))[0]
            ):
                if meta[1]:
                    await ctx.send(
                        _(
                            "Heap-size must be less than your system RAM. "
                            "You currently have {ram_in_bytes} of RAM available."
                        ).format(ram_in_bytes=inline(sizeof_fmt(meta[0])))
                    )
                else:
                    await ctx.send(
                        _(
                            "Heap-size must be less than {limit} due to your system limitations."
                        ).format(limit=inline(sizeof_fmt(meta[0])))
                    )
                return 0
            return 1

        if not (await validate_input(self, size)):
            return
        size = size.upper()
        await self.config.java.Xmx.set(size)
        await self.send_embed_msg(
            ctx,
            title=_("Setting Changed"),
            description=_(
                "Managed node's heap-size set to {bytes}.\n\n"
                "Run `{p}{cmd}` for it to take effect."
            ).format(
                bytes=inline(size), p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name
            ),
        )

    @command_llset.command(name="unmanaged", aliases=["external"])
    async def command_llset_unmanaged(self, ctx: commands.Context):
        """Toggle using external (unmanaged) Lavalink nodes - requires an existing Lavalink node for Audio to work, if enabled.

        This command disables the managed Lavalink server. If you do not have another Lavalink node set up, you will be unable to use Audio while this is enabled.
        """
        external = await self.config.use_external_lavalink()
        await self.config.use_external_lavalink.set(not external)
        async with ctx.typing():
            if external:
                embed = discord.Embed(
                    title=_("Setting Changed"),
                    description=_("Unmanaged Lavalink server: {true_or_false}.").format(
                        true_or_false=inline(_("Enabled") if not external else _("Disabled"))
                    ),
                )
                await self.send_embed_msg(ctx, embed=embed)
            else:
                await self.send_embed_msg(
                    ctx,
                    title=_("Setting Changed"),
                    description=_("Unmanaged Lavalink server: {true_or_false}.").format(
                        true_or_false=inline(_("Enabled") if not external else _("Disabled"))
                    ),
                )
            try:
                await lavalink.close(self.bot)
                self.lavalink_restart_connect(manual=True)
            except ProcessLookupError:
                await self.send_embed_msg(
                    ctx,
                    title=_("Failed To Shutdown Lavalink"),
                    description=_("Please reload Audio (`{prefix}reload audio`).").format(
                        prefix=ctx.prefix
                    ),
                )

    @command_llset.command(name="host")
    @has_unmanaged_server()
    async def command_llset_host(
        self, ctx: commands.Context, host: str = DEFAULT_LAVALINK_SETTINGS["host"]
    ):
        """Set the Lavalink node host.

        This command sets the connection host which Audio will use to connect to an unmanaged Lavalink node.
        """
        await self.config.host.set(host)
        await self.send_embed_msg(
            ctx,
            title=_("Setting Changed"),
            description=_(
                "Unmanaged Lavalink node host set to {host}. "
                "Run `{p}{cmd}` for it to take effect."
            ).format(
                host=inline(host), p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name
            ),
        )

    @command_llset.command(name="password", aliases=["pass", "token"])
    @has_unmanaged_server()
    async def command_llset_password(
        self, ctx: commands.Context, *, password: str = DEFAULT_LAVALINK_SETTINGS["password"]
    ):
        """Set the Lavalink node password.

        This command sets the connection password which Audio will use to connect to an unmanaged Lavalink node.
        """

        await self.config.password.set(str(password))
        await self.send_embed_msg(
            ctx,
            title=_("Setting Changed"),
            description=_(
                "Unmanaged Lavalink node password set to {password}. "
                "Run `{p}{cmd}` for it to take effect."
            ).format(
                password=inline(password),
                p=ctx.prefix,
                cmd=self.command_audioset_restart.qualified_name,
            ),
        )

    @command_llset.command(name="port")
    @has_unmanaged_server()
    async def command_llset_wsport(
        self, ctx: commands.Context, port: int = DEFAULT_LAVALINK_SETTINGS["ws_port"]
    ):
        """Set the Lavalink node port.

        This command sets the connection port which Audio will use to connect to an unmanaged Lavalink node.
        Set port to -1 to disable the port and connect to the specified host via ports 80/443
        """
        if port < 0:
            port = None
        elif port > 65535:
            return await self.send_embed_msg(
                ctx,
                title=_("Setting Not Changed"),
                description=_("A port must be between 0 and 65535."),
            )
        await self.config.ws_port.set(port)
        await self.send_embed_msg(
            ctx,
            title=_("Setting Changed"),
            description=_(
                "Unmanaged Lavalink node port set to {port}. "
                "Run `{p}{cmd}` for it to take effect."
            ).format(
                port=inline(str(port)),
                p=ctx.prefix,
                cmd=self.command_audioset_restart.qualified_name,
            ),
        )

    @command_llset.command(name="secured", aliases=["wss"])
    @has_unmanaged_server()
    async def command_llset_secured(self, ctx: commands.Context):
        """Set the Lavalink node connection to secured.

        This toggle sets the connection type to secured or unsecured when connecting to an unmanaged Lavalink node.
        """
        state = await self.config.secured_ws()
        await self.config.secured_ws.set(not state)

        if not state:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Unmanaged Lavalink node will now connect using the secured {secured_protocol} protocol.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(
                    p=ctx.prefix,
                    cmd=self.command_audioset_restart.qualified_name,
                    secured_protocol=inline("wss://"),
                ),
            )
        else:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Unmanaged Lavalink node will no longer connect using the secured "
                    "{secured_protocol} protocol and will use {unsecured_protocol} instead.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(
                    p=ctx.prefix,
                    cmd=self.command_audioset_restart.qualified_name,
                    unsecured_protocol=inline("ws://"),
                    secured_protocol=inline("wss://"),
                ),
            )

    @command_llset.command(name="info", aliases=["settings"])
    async def command_llset_info(self, ctx: commands.Context):
        """Display Lavalink connection settings."""
        configs = await self.config.all()

        if configs["use_external_lavalink"]:
            msg = "----" + _("Connection Settings") + "----        \n"
            msg += _("Host:             [{host}]\n").format(host=configs["host"])
            msg += _("Port:             [{port}]\n").format(
                port=configs["ws_port"] or _("Default HTTP/HTTPS port")
            )
            msg += _("Password:         [{password}]\n").format(password=configs["password"])
            msg += _("Secured:          [{state}]\n").format(state=configs["secured_ws"])

        else:
            msg = "----" + _("Lavalink Node Settings") + "----        \n"
            msg += _("Host:             [{host}]\n").format(
                host=configs["yaml"]["server"]["address"]
            )
            msg += _("Port:             [{port}]\n").format(port=configs["yaml"]["server"]["port"])
            msg += _("Password:         [{password}]\n").format(
                password=configs["yaml"]["lavalink"]["server"]["password"]
            )
            msg += _("Initial Heapsize: [{xms}]\n").format(xms=configs["java"]["Xms"])
            msg += _("Max Heapsize:     [{xmx}]\n").format(xmx=configs["java"]["Xmx"])
            msg += _("Java exec:        [{java_exc_path}]\n").format(
                java_exc_path=configs["java_exc_path"]
            )

        try:
            await self.send_embed_msg(ctx.author, description=box(msg, lang="ini"))
            await ctx.tick()
        except discord.HTTPException:
            await ctx.send(_("I need to be able to DM you to send you this info."))

    @command_llset.command(name="yaml", aliases=["yml"])
    @has_managed_server()
    async def command_llset_yaml(self, ctx: commands.Context):
        """Uploads a copy of the application.yml file used by the managed Lavalink node."""
        configs = change_dict_naming_convention(await self.config.yaml.all())
        data = yaml.safe_dump(configs)
        playlist_data = data.encode("utf-8")
        to_write = BytesIO()
        to_write.write(playlist_data)
        to_write.seek(0)
        datapath = cog_data_path(raw_name="Audio")
        temp_file = datapath / f"application.dump.yaml"
        try:
            with temp_file.open("wb") as application_file:
                application_file.write(to_write.read())
            await ctx.author.send(
                file=discord.File(str(temp_file)),
            )
            await ctx.tick()
        except discord.HTTPException:
            await ctx.send(_("I need to be able to DM you to send you this info."))
        finally:
            temp_file.unlink()

    @command_llset.group(name="config", aliases=["conf"])
    @has_managed_server()
    async def command_llset_config(self, ctx: commands.Context):
        """Configure the managed Lavalink node runtime options.

        All settings under this group will likely cause Audio to malfunction if changed from their defaults, only change settings here if you have been advised to by support.
        """

    @command_llset_config.group(name="server")
    async def command_llset_config_server(self, ctx: commands.Context):
        """Configure the managed node authorization and connection settings."""

    @command_llset_config.command(name="bind", aliases=["host", "address"])
    async def command_llset_config_host(
        self, ctx: commands.Context, *, host: str = DEFAULT_LAVALINK_YAML["yaml__server__address"]
    ):
        """`Dangerous command` Set the managed Lavalink node's binding IP address.

        This value by default is `localhost` which will restrict the server to only localhost apps by default, changing this will likely break the managed Lavalink node if you don't know what you are doing.
        """

        await self.config.yaml.server.address.set(host)
        await self.send_embed_msg(
            ctx,
            title=_("Setting Changed"),
            description=_(
                "Managed node will now accept connection on {host}.\n\n"
                "Run `{p}{cmd}` for it to take effect."
            ).format(
                host=inline(host), p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name
            ),
        )

    @command_llset_config.command(name="token", aliases=["password", "pass"])
    async def command_llset_config_token(
        self,
        ctx: commands.Context,
        *,
        password: str = DEFAULT_LAVALINK_YAML["yaml__lavalink__server__password"],
    ):
        """Set the managed Lavalink node's connection password.

        This is the password required for Audio to connect to the managed Lavalink node.
        The value by default is `youshallnotpass`.
        """
        await self.config.yaml.lavalink.server.password.set(password)
        await self.send_embed_msg(
            ctx,
            title=_("Setting Changed"),
            description=_(
                "Managed node will now accept {password} as the authorization token.\n\n"
                "Run `{p}{cmd}` for it to take effect."
            ).format(
                password=inline(password),
                p=ctx.prefix,
                cmd=self.command_audioset_restart.qualified_name,
            ),
        )

    @command_llset_config.command(name="port")
    async def command_llset_config_port(
        self, ctx: commands.Context, *, port: int = DEFAULT_LAVALINK_YAML["yaml__server__port"]
    ):
        """`Dangerous command` Set the managed Lavalink node's connection port.

        This port is the port the managed Lavalink node binds to, you should only change this if there is a conflict with the default port because you already have an application using port 2333 on this device.

        The value by default is `2333`.
        """
        if 1024 > port or port > 49151:
            return await self.send_embed_msg(
                ctx,
                title=_("Setting Not Changed"),
                description=_("The port must be between 1024 and 49151."),
            )

        await self.config.yaml.server.port.set(port)
        await self.send_embed_msg(
            ctx,
            title=_("Setting Changed"),
            description=_(
                "Managed node will now accept connections on {port}.\n\n"
                "Run `{p}{cmd}` for it to take effect."
            ).format(
                port=inline(str(port)),
                p=ctx.prefix,
                cmd=self.command_audioset_restart.qualified_name,
            ),
        )

    @command_llset_config.group(name="source")
    async def command_llset_config_source(self, ctx: commands.Context):
        """`Dangerous command` Toggle audio sources on/off.

        By default, all sources are enabled, you should only use commands here to disable a specific source if you have been advised to, disabling sources without background knowledge can cause Audio to break.
        """

    @command_llset_config_source.command(name="http")
    async def command_llset_config_source_http(self, ctx: commands.Context):
        """Toggle HTTP direct URL usage on or off.

        This source is used to allow playback from direct HTTP streams (this does not affect direct URL playback for the other sources).
        """
        state = await self.config.yaml.lavalink.server.sources.http()
        await self.config.yaml.lavalink.server.sources.http.set(not state)
        if not state:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will allow playback from direct URLs.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )
        else:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will not play from direct URLs anymore.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )

    @command_llset_config_source.command(name="bandcamp", aliases=["bc"])
    async def command_llset_config_source_bandcamp(self, ctx: commands.Context):
        """Toggle Bandcamp source on or off.

        This toggle controls the playback of all Bandcamp related content.
        """
        state = await self.config.yaml.bandcamp.http()
        await self.config.yaml.lavalink.server.sources.bandcamp.set(not state)
        if not state:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will allow playback from Bandcamp.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )
        else:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will not play from Bandcamp anymore.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )

    @command_llset_config_source.command(name="local")
    async def command_llset_config_source_local(self, ctx: commands.Context):
        """Toggle local file usage on or off.

        This toggle controls the playback of all local track content, usually found inside the `localtracks` folder.
        """
        state = await self.config.yaml.lavalink.server.sources.local()
        await self.config.yaml.lavalink.server.sources.local.set(not state)
        if not state:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will allow playback from local files.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )
        else:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will not play from local files anymore.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )

    @command_llset_config_source.command(name="soundcloud", aliases=["sc"])
    async def command_llset_config_source_soundcloud(self, ctx: commands.Context):
        """Toggle Soundcloud source on or off.

        This toggle controls the playback of all SoundCloud related content.
        """
        state = await self.config.yaml.lavalink.server.sources.soundcloud()
        await self.config.yaml.lavalink.server.sources.soundcloud.set(not state)
        if not state:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will allow playback from Soundcloud.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )
        else:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will not play from Soundcloud anymore.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )

    @command_llset_config_source.command(name="youtube", aliases=["yt"])
    async def command_llset_config_source_youtube(self, ctx: commands.Context):
        """`Dangerous command` Toggle YouTube source on or off (this includes Spotify).

        This toggle controls the playback of all YouTube and Spotify related content.
        """
        state = await self.config.yaml.lavalink.server.sources.youtube()
        await self.config.yaml.lavalink.server.sources.youtube.set(not state)
        if not state:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will allow playback from YouTube.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )
        else:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will not play from YouTube anymore.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )

    @command_llset_config_source.command(name="twitch")
    async def command_llset_config_source_twitch(self, ctx: commands.Context):
        """Toggle Twitch source on or off.

        This toggle controls the playback of all Twitch related content.
        """
        state = await self.config.yaml.lavalink.server.sources.twitch()
        await self.config.yaml.lavalink.server.sources.twitch.set(not state)
        if not state:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will allow playback from Twitch.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )
        else:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will not play from Twitch anymore.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )

    @command_llset_config_source.command(name="vimeo")
    async def command_llset_config_source_vimeo(self, ctx: commands.Context):
        """Toggle Vimeo source on or off.

        This toggle controls the playback of all Vimeo related content.
        """
        state = await self.config.yaml.lavalink.server.sources.vimeo()
        await self.config.yaml.lavalink.server.sources.vimeo.set(not state)
        if not state:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will allow playback from Vimeo.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )
        else:
            await self.send_embed_msg(
                ctx,
                title=_("Setting Changed"),
                description=_(
                    "Managed node will not play from Vimeo anymore.\n\n"
                    "Run `{p}{cmd}` for it to take effect."
                ).format(p=ctx.prefix, cmd=self.command_audioset_restart.qualified_name),
            )

    @command_llset_config_server.command(name="framebuffer", aliases=["fb", "frame"])
    async def command_llset_config_server_framebuffer(
        self,
        ctx: commands.Context,
        *,
        milliseconds: int = DEFAULT_LAVALINK_YAML["yaml__lavalink__server__frameBufferDurationMs"],
    ):
        """`Dangerous command` Set the managed Lavalink node framebuffer size.

        Only change this if you have been directly advised to, changing it can cause significant playback issues.
        """
        if milliseconds < 100:
            return await self.send_embed_msg(
                ctx,
                title=_("Setting Not Changed"),
                description=_("The lowest value the framebuffer can be set to is 100ms."),
            )
        await self.config.yaml.lavalink.server.frameBufferDurationMs.set(milliseconds)
        await self.send_embed_msg(
            ctx,
            title=_("Setting Changed"),
            description=_(
                "Managed node's bufferDurationMs set to {milliseconds}.\n\n"
                "Run `{p}{cmd}` for it to take effect."
            ).format(
                milliseconds=inline(str(milliseconds)),
                p=ctx.prefix,
                cmd=self.command_audioset_restart.qualified_name,
            ),
        )

    @command_llset_config_server.command(name="buffer", aliases=["b"])
    async def command_llset_config_server_buffer(
        self,
        ctx: commands.Context,
        *,
        milliseconds: int = DEFAULT_LAVALINK_YAML["yaml__lavalink__server__bufferDurationMs"],
    ):
        """`Dangerous command`  Set the managed Lavalink node JDA-NAS buffer size.

        Only change this if you have been directly advised to, changing it can cause significant playback issues.
        """
        if milliseconds < 100:
            return await self.send_embed_msg(
                ctx,
                title=_("Setting Not Changed"),
                description=_("The lowest value the buffer may be is 100ms."),
            )
        await self.config.yaml.lavalink.server.bufferDurationMs.set(milliseconds)
        await self.send_embed_msg(
            ctx,
            title=_("Setting Changed"),
            description=_(
                "Managed node's bufferDurationMs set to {milliseconds}.\n\n"
                "Run `{p}{cmd}` for it to take effect."
            ).format(
                milliseconds=inline(str(milliseconds)),
                p=ctx.prefix,
                cmd=self.command_audioset_restart.qualified_name,
            ),
        )

    @command_llset.command(name="reset")
    async def command_llset_reset(self, ctx: commands.Context):
        """Reset all `llset` changes back to their default values."""
        async with ctx.typing():
            async with self.config.all() as global_data:
                del global_data["yaml"]
                for key in (*DEFAULT_LAVALINK_SETTINGS.keys(), *DEFAULT_LAVALINK_YAML.keys()):
                    if key in global_data:
                        del global_data[key]
                del global_data["java"]
                del global_data["java_exc_path"]
                global_data["use_external_lavalink"] = False

            try:
                await lavalink.close(self.bot)
                self.lavalink_restart_connect(manual=True)
            except ProcessLookupError:
                await self.send_embed_msg(
                    ctx,
                    title=_("Failed To Shutdown Lavalink Node"),
                    description=_("Please reload Audio (`{prefix}reload audio`).").format(
                        prefix=ctx.prefix
                    ),
                )



import os
import sys
from pathlib import Path

from setuptools import find_namespace_packages, setup

ROOT_FOLDER = Path(__file__).parent.absolute()
REQUIREMENTS_FOLDER = ROOT_FOLDER / "requirements"

# Since we're importing `redbot` package, we have to ensure that it's in sys.path.
sys.path.insert(0, str(ROOT_FOLDER))

from redbot import VersionInfo

version, _ = VersionInfo._get_version(ignore_installed=True)


def get_requirements(fp):
    return [
        line.strip()
        for line in fp.read().splitlines()
        if line.strip() and not line.strip().startswith("#")
    ]


def extras_combined(*extra_names):
    return list(
        {
            req
            for extra_name, extra_reqs in extras_require.items()
            if not extra_names or extra_name in extra_names
            for req in extra_reqs
        }
    )


with open(REQUIREMENTS_FOLDER / "base.txt", encoding="utf-8") as fp:
    install_requires = get_requirements(fp)

extras_require = {}
for file in REQUIREMENTS_FOLDER.glob("extra-*.txt"):
    with file.open(encoding="utf-8") as fp:
        extras_require[file.stem[len("extra-") :]] = get_requirements(fp)

extras_require["dev"] = extras_combined()
extras_require["all"] = extras_combined("postgres")


python_requires = ">=3.8.1"
if not os.getenv("TOX_RED", False) or sys.version_info < (3, 12):
    python_requires += ",<3.12"

# Metadata and options defined in pyproject.toml
setup(
    version=version,
    python_requires=python_requires,
    # TODO: use [tool.setuptools.dynamic] table once this feature gets out of beta
    install_requires=install_requires,
    extras_require=extras_require,
    # TODO: use [project] table once PEP 639 gets accepted
    license_files=["LICENSE", "redbot/**/*.LICENSE"],
    # TODO: use [tool.setuptools.packages] table once this feature gets out of beta
    packages=find_namespace_packages(include=["redbot", "redbot.*"]),
)



from redbot import _early_init

# this needs to be called as early as possible
_early_init()

import asyncio
import functools
import getpass
import json
import logging
import os
import pip
import platform
import shutil
import signal
import sys
from argparse import Namespace
from copy import deepcopy
from pathlib import Path
from typing import Any, Awaitable, Callable, NoReturn, Optional, Union

import discord
import rich

import redbot.logging
from redbot import __version__
from redbot.core.bot import Red, ExitCodes, _NoOwnerSet
from redbot.core._cli import interactive_config, confirm, parse_cli_flags
from redbot.setup import get_data_dir, get_name, save_config
from redbot.core import data_manager, _drivers
from redbot.core._debuginfo import DebugInfo
from redbot.core._sharedlibdeprecation import SharedLibImportWarner


log = logging.getLogger("red.main")

#
#               Red - Discord Bot v3
#
#         Made by Twentysix, improved by many
#


def _get_instance_names():
    with data_manager.config_file.open(encoding="utf-8") as fs:
        data = json.load(fs)
    return sorted(data.keys())


def list_instances():
    if not data_manager.config_file.exists():
        print(
            "No instances have been configured! Configure one "
            "using `redbot-setup` before trying to run the bot!"
        )
        sys.exit(ExitCodes.CONFIGURATION_ERROR)
    else:
        text = "Configured Instances:\n\n"
        for instance_name in _get_instance_names():
            text += "{}\n".format(instance_name)
        print(text)
        sys.exit(ExitCodes.SHUTDOWN)


async def debug_info(red: Optional[Red] = None, *args: Any) -> None:
    """Shows debug information useful for debugging."""
    print(await DebugInfo(red).get_cli_text())


async def edit_instance(red, cli_flags):
    no_prompt = cli_flags.no_prompt
    token = cli_flags.token
    owner = cli_flags.owner
    prefix = cli_flags.prefix
    old_name = cli_flags.instance_name
    new_name = cli_flags.edit_instance_name
    data_path = cli_flags.edit_data_path
    copy_data = cli_flags.copy_data
    confirm_overwrite = cli_flags.overwrite_existing_instance

    if data_path is None and copy_data:
        print("--copy-data can't be used without --edit-data-path argument")
        sys.exit(ExitCodes.INVALID_CLI_USAGE)
    if new_name is None and confirm_overwrite:
        print("--overwrite-existing-instance can't be used without --edit-instance-name argument")
        sys.exit(ExitCodes.INVALID_CLI_USAGE)
    if (
        no_prompt
        and all(to_change is None for to_change in (token, owner, new_name, data_path))
        and not prefix
    ):
        print(
            "No arguments to edit were provided."
            " Available arguments (check help for more information):"
            " --edit-instance-name, --edit-data-path, --copy-data, --owner, --token, --prefix"
        )
        sys.exit(ExitCodes.INVALID_CLI_USAGE)

    await _edit_token(red, token, no_prompt)
    await _edit_prefix(red, prefix, no_prompt)
    await _edit_owner(red, owner, no_prompt)

    data = deepcopy(data_manager.basic_config)
    name = _edit_instance_name(old_name, new_name, confirm_overwrite, no_prompt)
    _edit_data_path(data, name, data_path, copy_data, no_prompt)

    save_config(name, data)
    if old_name != name:
        save_config(old_name, {}, remove=True)


async def _edit_token(red, token, no_prompt):
    if token:
        if not len(token) >= 50:
            print(
                "The provided token doesn't look a valid Discord bot token."
                " Instance's token will remain unchanged.\n"
            )
            return
        await red._config.token.set(token)
    elif not no_prompt and confirm("Would you like to change instance's token?", default=False):
        await interactive_config(red, False, True, print_header=False)
        print("Token updated.\n")


async def _edit_prefix(red, prefix, no_prompt):
    if prefix:
        prefixes = sorted(prefix, reverse=True)
        await red._config.prefix.set(prefixes)
    elif not no_prompt and confirm("Would you like to change instance's prefixes?", default=False):
        print(
            "Enter the prefixes, separated by a space (please note "
            "that prefixes containing a space will need to be added with [p]set prefix)"
        )
        while True:
            prefixes = input("> ").strip().split()
            if not prefixes:
                print("You need to pass at least one prefix!")
                continue
            if any(prefix.startswith("/") for prefix in prefixes):
                print(
                    "Prefixes cannot start with '/', as it conflicts with Discord's slash commands."
                )
                continue
            prefixes = sorted(prefixes, reverse=True)
            await red._config.prefix.set(prefixes)
            print("Prefixes updated.\n")
            break


async def _edit_owner(red, owner, no_prompt):
    if owner:
        if not (15 <= len(str(owner)) <= 20):
            print(
                "The provided owner id doesn't look like a valid Discord user id."
                " Instance's owner will remain unchanged."
            )
            return
        await red._config.owner.set(owner)
    elif not no_prompt and confirm("Would you like to change instance's owner?", default=False):
        print(
            "Remember:\n"
            "ONLY the person who is hosting Red should be owner."
            " This has SERIOUS security implications."
            " The owner can access any data that is present on the host system.\n"
        )
        if confirm("Are you sure you want to change instance's owner?", default=False):
            print("Please enter a Discord user id for new owner:")
            while True:
                owner_id = input("> ").strip()
                if not (15 <= len(owner_id) <= 20 and owner_id.isdecimal()):
                    print("That doesn't look like a valid Discord user id.")
                    continue
                owner_id = int(owner_id)
                await red._config.owner.set(owner_id)
                print("Owner updated.")
                break
        else:
            print("Instance's owner will remain unchanged.")
        print()


def _edit_instance_name(old_name, new_name, confirm_overwrite, no_prompt):
    if new_name:
        name = new_name
        if name in _get_instance_names() and not confirm_overwrite:
            name = old_name
            print(
                "An instance with this name already exists.\n"
                "If you want to remove the existing instance and replace it with this one,"
                " run this command with --overwrite-existing-instance flag."
            )
    elif not no_prompt and confirm("Would you like to change the instance name?", default=False):
        name = get_name("")
        if name in _get_instance_names():
            print(
                "WARNING: An instance already exists with this name. "
                "Continuing will overwrite the existing instance config."
            )
            if not confirm(
                "Are you absolutely certain you want to continue with this instance name?",
                default=False,
            ):
                print("Instance name will remain unchanged.")
                name = old_name
            else:
                print("Instance name updated.")
        else:
            print("Instance name updated.")
        print()
    else:
        name = old_name
    return name


def _edit_data_path(data, instance_name, data_path, copy_data, no_prompt):
    # This modifies the passed dict.
    if data_path:
        new_path = Path(data_path)
        try:
            exists = new_path.exists()
        except OSError:
            print(
                "We were unable to check your chosen directory."
                " Provided path may contain an invalid character."
                " Data location will remain unchanged."
            )

        if not exists:
            try:
                new_path.mkdir(parents=True, exist_ok=True)
            except OSError:
                print(
                    "We were unable to create your chosen directory."
                    " Data location will remain unchanged."
                )
        data["DATA_PATH"] = data_path
        if copy_data and not _copy_data(data):
            print("Can't copy data to non-empty location. Data location will remain unchanged.")
            data["DATA_PATH"] = data_manager.basic_config["DATA_PATH"]
    elif not no_prompt and confirm("Would you like to change the data location?", default=False):
        data["DATA_PATH"] = get_data_dir(
            instance_name=instance_name, data_path=None, interactive=True
        )
        if confirm("Do you want to copy the data from old location?", default=True):
            if not _copy_data(data):
                print("Can't copy the data to non-empty location.")
                if not confirm("Do you still want to use the new data location?"):
                    data["DATA_PATH"] = data_manager.basic_config["DATA_PATH"]
                    print("Data location will remain unchanged.")
                    return
            print("Old data has been copied over to the new location.")
        print("Data location updated.")


def _copy_data(data):
    if Path(data["DATA_PATH"]).exists():
        if any(os.scandir(data["DATA_PATH"])):
            return False
        else:
            # this is needed because copytree doesn't work when destination folder exists
            # Python 3.8 has `dirs_exist_ok` option for that
            os.rmdir(data["DATA_PATH"])
    shutil.copytree(data_manager.basic_config["DATA_PATH"], data["DATA_PATH"])
    return True


def early_exit_runner(
    cli_flags: Namespace,
    func: Union[Callable[[], Awaitable[Any]], Callable[[Red, Namespace], Awaitable[Any]]],
) -> NoReturn:
    """
    This one exists to not log all the things like it's a full run of the bot.
    """
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        if not cli_flags.instance_name:
            loop.run_until_complete(func())
            sys.exit(ExitCodes.SHUTDOWN)
            return

        data_manager.load_basic_configuration(cli_flags.instance_name)
        red = Red(cli_flags=cli_flags, description="Red V3", dm_help=None)
        driver_cls = _drivers.get_driver_class()
        loop.run_until_complete(driver_cls.initialize(**data_manager.storage_details()))
        loop.run_until_complete(func(red, cli_flags))
        loop.run_until_complete(driver_cls.teardown())
    except (KeyboardInterrupt, EOFError):
        print("Aborted!")
    finally:
        # note: sleep is unnecessary since we're not making any network connections
        asyncio.set_event_loop(None)
        loop.stop()
        loop.close()

    sys.exit(ExitCodes.SHUTDOWN)


async def run_bot(red: Red, cli_flags: Namespace) -> None:
    """
    This runs the bot.

    Any shutdown which is a result of not being able to log in needs to raise
    a SystemExit exception.

    If the bot starts normally, the bot should be left to handle the exit case.
    It will raise SystemExit in a task, which will reach the event loop and
    interrupt running forever, then trigger our cleanup process, and does not
    need additional handling in this function.
    """

    driver_cls = _drivers.get_driver_class()

    await driver_cls.initialize(**data_manager.storage_details())

    redbot.logging.init_logging(
        level=cli_flags.logging_level,
        location=data_manager.core_data_path() / "logs",
        cli_flags=cli_flags,
    )

    log.debug("====Basic Config====")
    log.debug("Data Path: %s", data_manager._base_data_path())
    log.debug("Storage Type: %s", data_manager.storage_type())

    # lib folder has to be in sys.path before trying to load any 3rd-party cog (GH-3061)
    # We might want to change handling of requirements in Downloader at later date
    LIB_PATH = data_manager.cog_data_path(raw_name="Downloader") / "lib"
    LIB_PATH.mkdir(parents=True, exist_ok=True)
    if str(LIB_PATH) not in sys.path:
        sys.path.append(str(LIB_PATH))

        # "It's important to note that the global `working_set` object is initialized from
        # `sys.path` when `pkg_resources` is first imported, but is only updated if you do
        # all future `sys.path` manipulation via `pkg_resources` APIs. If you manually modify
        # `sys.path`, you must invoke the appropriate methods on the `working_set` instance
        # to keep it in sync."
        # Source: https://setuptools.readthedocs.io/en/latest/pkg_resources.html#workingset-objects
        pkg_resources = sys.modules.get("pkg_resources")
        if pkg_resources is not None:
            pkg_resources.working_set.add_entry(str(LIB_PATH))
    sys.meta_path.insert(0, SharedLibImportWarner())

    if cli_flags.token:
        token = cli_flags.token
    else:
        token = os.environ.get("RED_TOKEN", None)
        if not token:
            token = await red._config.token()

    prefix = cli_flags.prefix or await red._config.prefix()

    if not (token and prefix):
        if cli_flags.no_prompt is False:
            new_token = await interactive_config(
                red, token_set=bool(token), prefix_set=bool(prefix)
            )
            if new_token:
                token = new_token
        else:
            log.critical("Token and prefix must be set in order to login.")
            sys.exit(ExitCodes.CONFIGURATION_ERROR)

    if cli_flags.dry_run:
        sys.exit(ExitCodes.SHUTDOWN)
    try:
        # `async with red:` is unnecessary here because we call red.close() in shutdown handler
        await red.start(token)
    except discord.LoginFailure:
        log.critical("This token doesn't seem to be valid.")
        db_token = await red._config.token()
        if db_token and not cli_flags.no_prompt:
            if confirm("\nDo you want to reset the token?"):
                await red._config.token.set("")
                print("Token has been reset.")
                sys.exit(ExitCodes.SHUTDOWN)
        sys.exit(ExitCodes.CONFIGURATION_ERROR)
    except discord.PrivilegedIntentsRequired:
        console = rich.get_console()
        console.print(
            "Red requires all Privileged Intents to be enabled.\n"
            "You can find out how to enable Privileged Intents with this guide:\n"
            "https://docs.discord.red/en/stable/bot_application_guide.html#enabling-privileged-intents",
            style="red",
        )
        sys.exit(ExitCodes.CONFIGURATION_ERROR)
    except _NoOwnerSet:
        print(
            "Bot doesn't have any owner set!\n"
            "This can happen when your bot's application is owned by team"
            " as team members are NOT owners by default.\n\n"
            "Remember:\n"
            "ONLY the person who is hosting Red should be owner."
            " This has SERIOUS security implications."
            " The owner can access any data that is present on the host system.\n"
            "With that out of the way, depending on who you want to be considered as owner,"
            " you can:\n"
            "a) pass --team-members-are-owners when launching Red"
            " - in this case Red will treat members of the bot application's team as owners,"
            " if their team role is Owner, Admin, or Developer\n"
            f"b) set owner manually with `redbot --edit {cli_flags.instance_name}`\n"
            "c) pass owner ID(s) when launching Red with --owner"
            " (and --co-owner if you need more than one) flag\n"
        )
        sys.exit(ExitCodes.CONFIGURATION_ERROR)

    return None


def handle_early_exit_flags(cli_flags: Namespace):
    if cli_flags.list_instances:
        list_instances()
    elif cli_flags.version:
        print("Red V3")
        print("Current Version: {}".format(__version__))
        sys.exit(ExitCodes.SHUTDOWN)
    elif cli_flags.debuginfo:
        early_exit_runner(cli_flags, debug_info)
    elif not cli_flags.instance_name and (not cli_flags.no_instance or cli_flags.edit):
        print("Error: No instance name was provided!")
        sys.exit(ExitCodes.INVALID_CLI_USAGE)


async def signal_shutdown_handler(red: Red, signal_type: signal.Signals) -> NoReturn:
    log.info("%s received. Quitting...", signal_type.name)
    sys.exit(ExitCodes.SHUTDOWN)


async def shutdown_handler(red: Red, exit_code: int) -> None:
    red._shutdown_mode = exit_code
    try:
        if not red.is_closed():
            await red.close()
    finally:
        # Then cancels all outstanding tasks other than ourselves
        pending = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        [task.cancel() for task in pending]
        await asyncio.gather(*pending, return_exceptions=True)


def global_exception_handler(red, loop, context):
    """
    Logs unhandled exceptions in other tasks
    """
    exc = context.get("exception")
    # These will get handled later when it *also* kills loop.run_forever
    if exc is not None and isinstance(exc, (KeyboardInterrupt, SystemExit)):
        return
    loop.default_exception_handler(context)


def red_exception_handler(red, red_task: asyncio.Future):
    """
    This is set as a done callback for Red

    must be used with functools.partial

    If the main bot.run dies for some reason,
    we don't want to swallow the exception and hang.
    """
    try:
        red_task.result()
    except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
        pass  # Handled by the global_exception_handler, or cancellation
    except Exception as exc:
        log.critical("The main bot task didn't handle an exception and has crashed", exc_info=exc)
        log.warning("Attempting to die as gracefully as possible...")
        log.info("Shutting down from unhandled exception")
        sys.exit(ExitCodes.CRITICAL)


def main():
    red = None  # Error handling for users misusing the bot
    cli_flags = parse_cli_flags(sys.argv[1:])
    handle_early_exit_flags(cli_flags)
    if cli_flags.edit:
        early_exit_runner(cli_flags, edit_instance)
        return
    try:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        if cli_flags.no_instance:
            print(
                "\033[1m"
                "Warning: The data will be placed in a temporary folder and removed on next system "
                "reboot."
                "\033[0m"
            )
            cli_flags.instance_name = "temporary_red"
            data_manager.create_temp_config()

        data_manager.load_basic_configuration(cli_flags.instance_name)

        red = Red(cli_flags=cli_flags, description="Red V3", dm_help=None)

        if os.name != "nt":
            # None of this works on windows.
            # At least it's not a redundant handler...
            signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
            for s in signals:
                loop.add_signal_handler(
                    s, lambda s=s: asyncio.create_task(signal_shutdown_handler(red, s))
                )

        exc_handler = functools.partial(global_exception_handler, red)
        loop.set_exception_handler(exc_handler)
        # We actually can't (just) use asyncio.run here
        # We probably could if we didn't support windows, but we might run into
        # a scenario where this isn't true if anyone works on RPC more in the future
        fut = loop.create_task(run_bot(red, cli_flags))
        r_exc_handler = functools.partial(red_exception_handler, red)
        fut.add_done_callback(r_exc_handler)
        loop.run_forever()
    except KeyboardInterrupt:
        # We still have to catch this here too. (*joy*)
        log.warning("Please do not use Ctrl+C to Shutdown Red! (attempting to die gracefully...)")
        log.error("Received KeyboardInterrupt, treating as interrupt")
        if red is not None:
            loop.run_until_complete(signal_shutdown_handler(red, signal.SIGINT))
    except SystemExit as exc:
        # We also have to catch this one here. Basically any exception which normally
        # Kills the python interpreter (Base Exceptions minus asyncio.cancelled)
        # We need to do something with prior to having the loop close
        exit_code = int(exc.code)
        try:
            exit_code_name = ExitCodes(exit_code).name
        except ValueError:
            exit_code_name = "UNKNOWN"
        log.info("Shutting down with exit code: %s (%s)", exit_code, exit_code_name)
        if red is not None:
            loop.run_until_complete(shutdown_handler(red, exc.code))
    except Exception as exc:  # Non standard case.
        log.exception("Unexpected exception (%s): ", type(exc), exc_info=exc)
        if red is not None:
            loop.run_until_complete(shutdown_handler(red, ExitCodes.CRITICAL))
    finally:
        # Allows transports to close properly, and prevent new ones from being opened.
        # Transports may still not be closed correctly on windows, see below
        loop.run_until_complete(loop.shutdown_asyncgens())
        # *we* aren't cleaning up more here, but it prevents
        # a runtime error at the event loop on windows
        # with resources which require longer to clean up.
        # With other event loops, a failure to cleanup prior to here
        # results in a resource warning instead
        log.info("Please wait, cleaning up a bit more")
        loop.run_until_complete(asyncio.sleep(2))
        asyncio.set_event_loop(None)
        loop.stop()
        loop.close()
    exit_code = red._shutdown_mode if red is not None else ExitCodes.CRITICAL
    sys.exit(exit_code)


if __name__ == "__main__":
    main()



import contextlib
import math
import time
from pathlib import Path

from typing import MutableMapping

import discord
import lavalink
from red_commons.logging import getLogger

from lavalink import NodeNotFound

from redbot.core import commands
from redbot.core.commands import UserInputOptional
from redbot.core.i18n import Translator
from redbot.core.utils import AsyncIter
from redbot.core.utils.menus import close_menu, menu, next_page, prev_page

from ...audio_dataclasses import _PARTIALLY_SUPPORTED_MUSIC_EXT, Query
from ...errors import (
    DatabaseError,
    QueryUnauthorized,
    SpotifyFetchError,
    TrackEnqueueError,
)
from ..abc import MixinMeta
from ..cog_utils import CompositeMetaClass

log = getLogger("red.cogs.Audio.cog.Commands.player")
_ = Translator("Audio", Path(__file__))


class PlayerCommands(MixinMeta, metaclass=CompositeMetaClass):
    @commands.command(name="play")
    @commands.guild_only()
    @commands.bot_has_permissions(embed_links=True)
    async def command_play(self, ctx: commands.Context, *, query: str):
        """Play the specified track or search for a close match.

        To play a local track, the query should be `<parentfolder>\\<filename>`.
        If you are the bot owner, use `[p]audioset info` to display your localtracks path.
        """
        query = Query.process_input(query, self.local_folder_current_path)
        guild_data = await self.config.guild(ctx.guild).all()
        restrict = await self.config.restrict()
        if restrict and self.match_url(str(query)):
            valid_url = self.is_url_allowed(str(query))
            if not valid_url:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_(
                        "That URL is not allowed.\n\nThe bot owner can remove this restriction by using ``{prefix}audioset restrict``."
                    ).format(prefix=ctx.clean_prefix),
                )
        elif not await self.is_query_allowed(self.config, ctx, f"{query}", query_obj=query):
            return await self.send_embed_msg(
                ctx, title=_("Unable To Play Tracks"), description=_("That track is not allowed.")
            )
        can_skip = await self._can_instaskip(ctx, ctx.author)
        if guild_data["dj_enabled"] and not can_skip:
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("You need the DJ role to queue tracks."),
            )
        if not self._player_check(ctx):
            if self.lavalink_connection_aborted:
                msg = _("Connection to Lavalink node has failed")
                desc = None
                if await self.bot.is_owner(ctx.author):
                    desc = _("Please check your console or logs for details.")
                return await self.send_embed_msg(ctx, title=msg, description=desc)
            try:
                if (
                    not self.can_join_and_speak(ctx.author.voice.channel)
                    or not ctx.author.voice.channel.permissions_for(ctx.me).move_members
                    and self.is_vc_full(ctx.author.voice.channel)
                ):
                    return await self.send_embed_msg(
                        ctx,
                        title=_("Unable To Play Tracks"),
                        description=_(
                            "I don't have permission to connect and speak in your channel."
                        ),
                    )
                await lavalink.connect(
                    ctx.author.voice.channel,
                    self_deaf=await self.config.guild_from_id(ctx.guild.id).auto_deafen(),
                )
            except AttributeError:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("Connect to a voice channel first."),
                )
            except NodeNotFound:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("Connection to the Lavalink node has not yet been established."),
                )
        player = lavalink.get_player(ctx.guild.id)
        player.store("notify_channel", ctx.channel.id)
        await self._eq_check(ctx, player)
        await self.set_player_settings(ctx)
        if (not ctx.author.voice or ctx.author.voice.channel != player.channel) and not can_skip:
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("You must be in the voice channel to use the play command."),
            )
        if not query.valid:
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("No tracks found for `{query}`.").format(
                    query=query.to_string_user()
                ),
            )
        if len(player.queue) >= 10000:
            return await self.send_embed_msg(
                ctx, title=_("Unable To Play Tracks"), description=_("Queue size limit reached.")
            )

        if not await self.maybe_charge_requester(ctx, guild_data["jukebox_price"]):
            return
        if query.is_spotify:
            return await self._get_spotify_tracks(ctx, query)
        try:
            await self._enqueue_tracks(ctx, query)
        except QueryUnauthorized as exc:
            return await self.send_embed_msg(
                ctx, title=_("Unable To Play Tracks"), description=exc.message
            )
        except Exception as e:
            self.update_player_lock(ctx, False)
            raise e

    @commands.command(name="bumpplay")
    @commands.guild_only()
    @commands.bot_has_permissions(embed_links=True)
    async def command_bumpplay(
        self, ctx: commands.Context, play_now: UserInputOptional[bool] = False, *, query: str
    ):
        """Force play a URL or search for a track."""
        query = Query.process_input(query, self.local_folder_current_path)
        if not query.single_track:
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Bump Track"),
                description=_("Only single tracks work with bump play."),
            )
        guild_data = await self.config.guild(ctx.guild).all()
        restrict = await self.config.restrict()
        if restrict and self.match_url(str(query)):
            valid_url = self.is_url_allowed(str(query))
            if not valid_url:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("That URL is not allowed."),
                )
        elif not await self.is_query_allowed(self.config, ctx, f"{query}", query_obj=query):
            return await self.send_embed_msg(
                ctx, title=_("Unable To Play Tracks"), description=_("That track is not allowed.")
            )
        can_skip = await self._can_instaskip(ctx, ctx.author)
        if guild_data["dj_enabled"] and not can_skip:
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("You need the DJ role to queue tracks."),
            )
        if not self._player_check(ctx):
            if self.lavalink_connection_aborted:
                msg = _("Connection to Lavalink node has failed")
                desc = None
                if await self.bot.is_owner(ctx.author):
                    desc = _("Please check your console or logs for details.")
                return await self.send_embed_msg(ctx, title=msg, description=desc)
            try:
                if (
                    not self.can_join_and_speak(ctx.author.voice.channel)
                    or not ctx.author.voice.channel.permissions_for(ctx.me).move_members
                    and self.is_vc_full(ctx.author.voice.channel)
                ):
                    return await self.send_embed_msg(
                        ctx,
                        title=_("Unable To Play Tracks"),
                        description=_(
                            "I don't have permission to connect and speak in your channel."
                        ),
                    )
                await lavalink.connect(
                    ctx.author.voice.channel,
                    self_deaf=await self.config.guild_from_id(ctx.guild.id).auto_deafen(),
                )
            except AttributeError:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("Connect to a voice channel first."),
                )
            except NodeNotFound:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("Connection to Lavalink node has not yet been established."),
                )
        player = lavalink.get_player(ctx.guild.id)
        player.store("notify_channel", ctx.channel.id)
        await self._eq_check(ctx, player)
        await self.set_player_settings(ctx)
        if (not ctx.author.voice or ctx.author.voice.channel != player.channel) and not can_skip:
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("You must be in the voice channel to use the play command."),
            )
        if not query.valid:
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("No tracks found for `{query}`.").format(
                    query=query.to_string_user()
                ),
            )
        if len(player.queue) >= 10000:
            return await self.send_embed_msg(
                ctx, title=_("Unable To Play Tracks"), description=_("Queue size limit reached.")
            )

        if not await self.maybe_charge_requester(ctx, guild_data["jukebox_price"]):
            return
        try:
            if query.is_spotify:
                tracks = await self._get_spotify_tracks(ctx, query)
            else:
                tracks = await self._enqueue_tracks(ctx, query, enqueue=False)
        except QueryUnauthorized as exc:
            return await self.send_embed_msg(
                ctx, title=_("Unable To Play Tracks"), description=exc.message
            )
        except Exception as e:
            self.update_player_lock(ctx, False)
            raise e
        if isinstance(tracks, discord.Message):
            return
        elif not tracks:
            self.update_player_lock(ctx, False)
            title = _("Unable To Play Tracks")
            desc = _("No tracks found for `{query}`.").format(query=query.to_string_user())
            embed = discord.Embed(title=title, description=desc)
            if await self.config.use_external_lavalink() and query.is_local:
                embed.description = _(
                    "Local tracks will not work "
                    "if the managed Lavalink node cannot see the track.\n"
                    "This may be due to permissions or you are using an external Lavalink node "
                    "in a different machine than the bot and the local tracks."
                )
            elif query.is_local and query.suffix in _PARTIALLY_SUPPORTED_MUSIC_EXT:
                title = _("Track is not playable.")
                embed = discord.Embed(title=title)
                embed.description = _(
                    "**{suffix}** is not a fully supported format and some tracks may not play."
                ).format(suffix=query.suffix)
            return await self.send_embed_msg(ctx, embed=embed)
        queue_dur = await self.track_remaining_duration(ctx)
        index = query.track_index
        seek = 0
        if query.start_time:
            seek = query.start_time
        single_track = (
            tracks
            if isinstance(tracks, lavalink.rest_api.Track)
            else tracks[index]
            if index
            else tracks[0]
        )
        if seek and seek > 0:
            single_track.start_timestamp = seek * 1000
        query = Query.process_input(single_track, self.local_folder_current_path)
        if not await self.is_query_allowed(
            self.config,
            ctx,
            f"{single_track.title} {single_track.author} {single_track.uri} {str(query)}",
            query_obj=query,
        ):
            log.debug("Query is not allowed in %r (%s)", ctx.guild.name, ctx.guild.id)
            self.update_player_lock(ctx, False)
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("This track is not allowed in this server."),
            )
        elif guild_data["maxlength"] > 0:
            if self.is_track_length_allowed(single_track, guild_data["maxlength"]):
                single_track.requester = ctx.author
                single_track.extras.update(
                    {
                        "enqueue_time": int(time.time()),
                        "vc": player.channel.id,
                        "requester": ctx.author.id,
                    }
                )
                player.queue.insert(0, single_track)
                player.maybe_shuffle()
                self.bot.dispatch(
                    "red_audio_track_enqueue", player.guild, single_track, ctx.author
                )
            else:
                self.update_player_lock(ctx, False)
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("Track exceeds maximum length."),
                )

        else:
            single_track.requester = ctx.author
            single_track.extras["bumped"] = True
            single_track.extras.update(
                {
                    "enqueue_time": int(time.time()),
                    "vc": player.channel.id,
                    "requester": ctx.author.id,
                }
            )
            player.queue.insert(0, single_track)
            player.maybe_shuffle()
            self.bot.dispatch("red_audio_track_enqueue", player.guild, single_track, ctx.author)
        description = await self.get_track_description(
            single_track, self.local_folder_current_path
        )
        footer = None
        if not play_now and not guild_data["shuffle"] and queue_dur > 0:
            footer = _("{time} until track playback: #1 in queue").format(
                time=self.format_time(queue_dur)
            )
        await self.send_embed_msg(
            ctx, title=_("Track Enqueued"), description=description, footer=footer
        )

        if not player.current:
            await player.play()
        elif play_now:
            await player.skip()

        self.update_player_lock(ctx, False)

    @commands.command(name="genre")
    @commands.guild_only()
    @commands.bot_has_permissions(embed_links=True)
    async def command_genre(self, ctx: commands.Context):
        """Pick a Spotify playlist from a list of categories to start playing."""

        async def _category_search_menu(
            ctx: commands.Context,
            pages: list,
            controls: MutableMapping,
            message: discord.Message,
            page: int,
            timeout: float,
            emoji: str,
        ):
            if message:
                output = await self._genre_search_button_action(ctx, category_list, emoji, page)
                with contextlib.suppress(discord.HTTPException):
                    await message.delete()
                return output

        async def _playlist_search_menu(
            ctx: commands.Context,
            pages: list,
            controls: MutableMapping,
            message: discord.Message,
            page: int,
            timeout: float,
            emoji: str,
        ):
            if message:
                output = await self._genre_search_button_action(
                    ctx, playlists_list, emoji, page, playlist=True
                )
                with contextlib.suppress(discord.HTTPException):
                    await message.delete()
                return output

        category_search_controls = {
            "\N{DIGIT ONE}\N{COMBINING ENCLOSING KEYCAP}": _category_search_menu,
            "\N{DIGIT TWO}\N{COMBINING ENCLOSING KEYCAP}": _category_search_menu,
            "\N{DIGIT THREE}\N{COMBINING ENCLOSING KEYCAP}": _category_search_menu,
            "\N{DIGIT FOUR}\N{COMBINING ENCLOSING KEYCAP}": _category_search_menu,
            "\N{DIGIT FIVE}\N{COMBINING ENCLOSING KEYCAP}": _category_search_menu,
            "\N{LEFTWARDS BLACK ARROW}\N{VARIATION SELECTOR-16}": prev_page,
            "\N{CROSS MARK}": close_menu,
            "\N{BLACK RIGHTWARDS ARROW}\N{VARIATION SELECTOR-16}": next_page,
        }
        playlist_search_controls = {
            "\N{DIGIT ONE}\N{COMBINING ENCLOSING KEYCAP}": _playlist_search_menu,
            "\N{DIGIT TWO}\N{COMBINING ENCLOSING KEYCAP}": _playlist_search_menu,
            "\N{DIGIT THREE}\N{COMBINING ENCLOSING KEYCAP}": _playlist_search_menu,
            "\N{DIGIT FOUR}\N{COMBINING ENCLOSING KEYCAP}": _playlist_search_menu,
            "\N{DIGIT FIVE}\N{COMBINING ENCLOSING KEYCAP}": _playlist_search_menu,
            "\N{LEFTWARDS BLACK ARROW}\N{VARIATION SELECTOR-16}": prev_page,
            "\N{CROSS MARK}": close_menu,
            "\N{BLACK RIGHTWARDS ARROW}\N{VARIATION SELECTOR-16}": next_page,
        }

        api_data = await self._check_api_tokens()
        if any([not api_data["spotify_client_id"], not api_data["spotify_client_secret"]]):
            return await self.send_embed_msg(
                ctx,
                title=_("Invalid Environment"),
                description=_(
                    "The owner needs to set the Spotify client ID and Spotify client secret, "
                    "before Spotify URLs or codes can be used. "
                    "\nSee `{prefix}audioset spotifyapi` for instructions."
                ).format(prefix=ctx.prefix),
            )
        elif not api_data["youtube_api"]:
            return await self.send_embed_msg(
                ctx,
                title=_("Invalid Environment"),
                description=_(
                    "The owner needs to set the YouTube API key before Spotify URLs or "
                    "codes can be used.\nSee `{prefix}audioset youtubeapi` for instructions."
                ).format(prefix=ctx.prefix),
            )
        guild_data = await self.config.guild(ctx.guild).all()
        if guild_data["dj_enabled"] and not await self._can_instaskip(ctx, ctx.author):
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("You need the DJ role to queue tracks."),
            )
        if not self._player_check(ctx):
            if self.lavalink_connection_aborted:
                msg = _("Connection to Lavalink node has failed")
                desc = None
                if await self.bot.is_owner(ctx.author):
                    desc = _("Please check your console or logs for details.")
                return await self.send_embed_msg(ctx, title=msg, description=desc)
            try:
                if (
                    not self.can_join_and_speak(ctx.author.voice.channel)
                    or not ctx.author.voice.channel.permissions_for(ctx.me).move_members
                    and self.is_vc_full(ctx.author.voice.channel)
                ):
                    return await self.send_embed_msg(
                        ctx,
                        title=_("Unable To Play Tracks"),
                        description=_(
                            "I don't have permission to connect and speak in your channel."
                        ),
                    )
                await lavalink.connect(
                    ctx.author.voice.channel,
                    self_deaf=await self.config.guild_from_id(ctx.guild.id).auto_deafen(),
                )
            except AttributeError:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("Connect to a voice channel first."),
                )
            except NodeNotFound:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("Connection to Lavalink node has not yet been established."),
                )
        player = lavalink.get_player(ctx.guild.id)
        player.store("notify_channel", ctx.channel.id)
        await self._eq_check(ctx, player)
        await self.set_player_settings(ctx)
        if (
            not ctx.author.voice or ctx.author.voice.channel != player.channel
        ) and not await self._can_instaskip(ctx, ctx.author):
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("You must be in the voice channel to use the genre command."),
            )
        try:
            category_list = await self.api_interface.spotify_api.get_categories(ctx=ctx)
        except SpotifyFetchError as error:
            return await self.send_embed_msg(
                ctx,
                title=_("No categories found"),
                description=error.message.format(prefix=ctx.prefix),
            )
        if not category_list:
            return await self.send_embed_msg(ctx, title=_("No categories found, try again later."))
        len_folder_pages = math.ceil(len(category_list) / 5)
        category_search_page_list = []
        async for page_num in AsyncIter(range(1, len_folder_pages + 1)):
            embed = await self._build_genre_search_page(
                ctx, category_list, page_num, _("Categories")
            )
            category_search_page_list.append(embed)
        cat_menu_output = await menu(ctx, category_search_page_list, category_search_controls)
        if not cat_menu_output:
            return await self.send_embed_msg(
                ctx, title=_("No categories selected, try again later.")
            )
        category_name, category_pick = cat_menu_output
        playlists_list = await self.api_interface.spotify_api.get_playlist_from_category(
            category_pick, ctx=ctx
        )
        if not playlists_list:
            return await self.send_embed_msg(ctx, title=_("No categories found, try again later."))
        len_folder_pages = math.ceil(len(playlists_list) / 5)
        playlists_search_page_list = []
        async for page_num in AsyncIter(range(1, len_folder_pages + 1)):
            embed = await self._build_genre_search_page(
                ctx,
                playlists_list,
                page_num,
                _("Playlists for {friendly_name}").format(friendly_name=category_name),
                playlist=True,
            )
            playlists_search_page_list.append(embed)
        playlists_pick = await menu(ctx, playlists_search_page_list, playlist_search_controls)
        query = Query.process_input(playlists_pick, self.local_folder_current_path)
        if not query.valid:
            return await self.send_embed_msg(ctx, title=_("No tracks to play."))
        if len(player.queue) >= 10000:
            return await self.send_embed_msg(
                ctx, title=_("Unable To Play Tracks"), description=_("Queue size limit reached.")
            )
        if not await self.maybe_charge_requester(ctx, guild_data["jukebox_price"]):
            return
        if query.is_spotify:
            return await self._get_spotify_tracks(ctx, query)
        return await self.send_embed_msg(
            ctx, title=_("Couldn't find tracks for the selected playlist.")
        )

    @commands.command(name="autoplay")
    @commands.guild_only()
    @commands.bot_has_permissions(embed_links=True)
    @commands.mod_or_permissions(manage_guild=True)
    async def command_autoplay(self, ctx: commands.Context):
        """Starts auto play."""
        guild_data = await self.config.guild(ctx.guild).all()
        if guild_data["dj_enabled"] and not await self._can_instaskip(ctx, ctx.author):
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("You need the DJ role to queue tracks."),
            )
        if not self._player_check(ctx):
            if self.lavalink_connection_aborted:
                msg = _("Connection to Lavalink node has failed")
                desc = None
                if await self.bot.is_owner(ctx.author):
                    desc = _("Please check your console or logs for details.")
                return await self.send_embed_msg(ctx, title=msg, description=desc)
            try:
                if (
                    not self.can_join_and_speak(ctx.author.voice.channel)
                    or not ctx.author.voice.channel.permissions_for(ctx.me).move_members
                    and self.is_vc_full(ctx.author.voice.channel)
                ):
                    return await self.send_embed_msg(
                        ctx,
                        title=_("Unable To Play Tracks"),
                        description=_(
                            "I don't have permission to connect and speak in your channel."
                        ),
                    )
                await lavalink.connect(
                    ctx.author.voice.channel,
                    self_deaf=await self.config.guild_from_id(ctx.guild.id).auto_deafen(),
                )
            except AttributeError:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("Connect to a voice channel first."),
                )
            except NodeNotFound:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("Connection to Lavalink node has not yet been established."),
                )
        player = lavalink.get_player(ctx.guild.id)
        player.store("notify_channel", ctx.channel.id)
        await self._eq_check(ctx, player)
        await self.set_player_settings(ctx)
        if (
            not ctx.author.voice or ctx.author.voice.channel != player.channel
        ) and not await self._can_instaskip(ctx, ctx.author):
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Play Tracks"),
                description=_("You must be in the voice channel to use the autoplay command."),
            )
        if len(player.queue) >= 10000:
            return await self.send_embed_msg(
                ctx, title=_("Unable To Play Tracks"), description=_("Queue size limit reached.")
            )
        if not await self.maybe_charge_requester(ctx, guild_data["jukebox_price"]):
            return
        try:
            await self.api_interface.autoplay(player, self.playlist_api)
        except DatabaseError:
            notify_channel = player.fetch("notify_channel")
            if notify_channel:
                notify_channel = ctx.guild.get_channel_or_thread(notify_channel)
                await self.send_embed_msg(notify_channel, title=_("Couldn't get a valid track."))
            return
        except TrackEnqueueError:
            self.update_player_lock(ctx, False)
            return await self.send_embed_msg(
                ctx,
                title=_("Unable to Get Track"),
                description=_(
                    "I'm unable to get a track from Lavalink at the moment, try again in a few "
                    "minutes."
                ),
            )
        except Exception as e:
            self.update_player_lock(ctx, False)
            raise e

        if not guild_data["auto_play"]:
            await ctx.invoke(self.command_audioset_autoplay_toggle)
        if not guild_data["notify"] and not player.fetch("autoplay_notified", False):
            pass
        elif player.current:
            await self.send_embed_msg(ctx, title=_("Adding a track to queue."))

    @commands.command(name="search")
    @commands.guild_only()
    @commands.bot_has_permissions(embed_links=True)
    @commands.bot_can_react()
    async def command_search(self, ctx: commands.Context, *, query: str):
        """Pick a track with a search.

        Use `[p]search list <search term>` to queue all tracks found on YouTube. Use `[p]search sc
        <search term>` to search on SoundCloud instead of YouTube.
        """

        if not isinstance(query, (str, list, Query)):
            raise RuntimeError(
                f"Expected 'query' to be a string, list or Query object but received: {type(query)} - this is an unexpected argument type, please report it."
            )

        async def _search_menu(
            ctx: commands.Context,
            pages: list,
            controls: MutableMapping,
            message: discord.Message,
            page: int,
            timeout: float,
            emoji: str,
        ):
            if message:
                await self._search_button_action(ctx, tracks, emoji, page)
                with contextlib.suppress(discord.HTTPException):
                    await message.delete()
                return None

        search_controls = {
            "\N{DIGIT ONE}\N{COMBINING ENCLOSING KEYCAP}": _search_menu,
            "\N{DIGIT TWO}\N{COMBINING ENCLOSING KEYCAP}": _search_menu,
            "\N{DIGIT THREE}\N{COMBINING ENCLOSING KEYCAP}": _search_menu,
            "\N{DIGIT FOUR}\N{COMBINING ENCLOSING KEYCAP}": _search_menu,
            "\N{DIGIT FIVE}\N{COMBINING ENCLOSING KEYCAP}": _search_menu,
            "\N{LEFTWARDS BLACK ARROW}\N{VARIATION SELECTOR-16}": prev_page,
            "\N{CROSS MARK}": close_menu,
            "\N{BLACK RIGHTWARDS ARROW}\N{VARIATION SELECTOR-16}": next_page,
        }

        if not self._player_check(ctx):
            if self.lavalink_connection_aborted:
                msg = _("Connection to Lavalink has failed")
                desc = None
                if await self.bot.is_owner(ctx.author):
                    desc = _("Please check your console or logs for details.")
                return await self.send_embed_msg(ctx, title=msg, description=desc)
            try:
                if (
                    not self.can_join_and_speak(ctx.author.voice.channel)
                    or not ctx.author.voice.channel.permissions_for(ctx.me).move_members
                    and self.is_vc_full(ctx.author.voice.channel)
                ):
                    return await self.send_embed_msg(
                        ctx,
                        title=_("Unable To Search For Tracks"),
                        description=_(
                            "I don't have permission to connect and speak in your channel."
                        ),
                    )
                await lavalink.connect(
                    ctx.author.voice.channel,
                    self_deaf=await self.config.guild_from_id(ctx.guild.id).auto_deafen(),
                )
            except AttributeError:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Search For Tracks"),
                    description=_("Connect to a voice channel first."),
                )
            except NodeNotFound:
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Search For Tracks"),
                    description=_("Connection to Lavalink node has not yet been established."),
                )
        player = lavalink.get_player(ctx.guild.id)
        guild_data = await self.config.guild(ctx.guild).all()
        player.store("notify_channel", ctx.channel.id)
        can_skip = await self._can_instaskip(ctx, ctx.author)
        if (not ctx.author.voice or ctx.author.voice.channel != player.channel) and not can_skip:
            return await self.send_embed_msg(
                ctx,
                title=_("Unable To Search For Tracks"),
                description=_("You must be in the voice channel to enqueue tracks."),
            )
        await self._eq_check(ctx, player)
        await self.set_player_settings(ctx)

        before_queue_length = len(player.queue)

        if not isinstance(query, list):
            query = Query.process_input(query, self.local_folder_current_path)
            restrict = await self.config.restrict()
            if restrict and self.match_url(str(query)):
                valid_url = self.is_url_allowed(str(query))
                if not valid_url:
                    return await self.send_embed_msg(
                        ctx,
                        title=_("Unable To Play Tracks"),
                        description=_("That URL is not allowed."),
                    )
            if not await self.is_query_allowed(self.config, ctx, f"{query}", query_obj=query):
                return await self.send_embed_msg(
                    ctx,
                    title=_("Unable To Play Tracks"),
                    description=_("That track is not allowed."),
                )
            if query.invoked_from == "search list" or query.invoked_from == "local folder":
                if query.invoked_from == "search list" and not query.is_local:
                    try:
                        result, called_api = await self.api_interface.fetch_track(
                            ctx, player, query
                        )
                    except TrackEnqueueError:
                        self.update_player_lock(ctx, False)
                        return await self.send_embed_msg(
                            ctx,
                            title=_("Unable to Get Track"),
                            description=_(
                                "I'm unable to get a track from Lavalink at the moment, "
                                "try again in a few minutes."
                            ),
                        )
                    except Exception as e:
                        self.update_player_lock(ctx, False)
                        raise e

                    tracks = result.tracks
                else:
                    try:
                        query.search_subfolders = True
                        tracks = await self.get_localtrack_folder_tracks(ctx, player, query)
                    except TrackEnqueueError:
                        self.update_player_lock(ctx, False)
                        return await self.send_embed_msg(
                            ctx,
                            title=_("Unable to Get Track"),
                            description=_(
                                "I'm unable to get a track from Lavalink at the moment, "
                                "try again in a few minutes."
                            ),
                        )
                    except Exception as e:
                        self.update_player_lock(ctx, False)
                        raise e
                if not tracks:
                    embed = discord.Embed(title=_("Nothing found."))
                    if await self.config.use_external_lavalink() and query.is_local:
                        embed.description = _(
                            "Local tracks will not work "
                            "if the `Lavalink.jar` cannot see the track.\n"
                            "This may be due to permissions or because Lavalink.jar is being run "
                            "in a different machine than the local tracks."
                        )
                    elif query.is_local and query.suffix in _PARTIALLY_SUPPORTED_MUSIC_EXT:
                        embed = discord.Embed(title=_("Track is not playable."))
                        embed.description = _(
                            "**{suffix}** is not a fully supported format and some "
                            "tracks may not play."
                        ).format(suffix=query.suffix)
                    return await self.send_embed_msg(ctx, embed=embed)
                queue_dur = await self.queue_duration(ctx)
                queue_total_duration = self.format_time(queue_dur)
                if guild_data["dj_enabled"] and not can_skip:
                    return await self.send_embed_msg(
                        ctx,
                        title=_("Unable To Play Tracks"),
                        description=_("You need the DJ role to queue tracks."),
                    )
                track_len = 0
                empty_queue = not player.queue
                async for track in AsyncIter(tracks):
                    if len(player.queue) >= 10000:
                        continue
                    query = Query.process_input(track, self.local_folder_current_path)
                    if not await self.is_query_allowed(
                        self.config,
                        ctx,
                        f"{track.title} {track.author} {track.uri} " f"{str(query)}",
                        query_obj=query,
                    ):
                        log.debug("Query is not allowed in %r (%s)", ctx.guild.name, ctx.guild.id)
                        continue
                    elif guild_data["maxlength"] > 0:
                        if self.is_track_length_allowed(track, guild_data["maxlength"]):
                            track_len += 1
                            track.extras.update(
                                {
                                    "enqueue_time": int(time.time()),
                                    "vc": player.channel.id,
                                    "requester": ctx.author.id,
                                }
                            )
                            player.add(ctx.author, track)
                            self.bot.dispatch(
                                "red_audio_track_enqueue", player.guild, track, ctx.author
                            )
                    else:
                        track_len += 1
                        track.extras.update(
                            {
                                "enqueue_time": int(time.time()),
                                "vc": player.channel.id,
                                "requester": ctx.author.id,
                            }
                        )
                        player.add(ctx.author, track)
                        self.bot.dispatch(
                            "red_audio_track_enqueue", player.guild, track, ctx.author
                        )
                    if not player.current:
                        await player.play()
                player.maybe_shuffle(0 if empty_queue else 1)
                if len(tracks) > track_len:
                    maxlength_msg = _(" {bad_tracks} tracks cannot be queued.").format(
                        bad_tracks=(len(tracks) - track_len)
                    )
                else:
                    maxlength_msg = ""
                songembed = discord.Embed(
                    title=_("Queued {num} track(s).{maxlength_msg}").format(
                        num=track_len, maxlength_msg=maxlength_msg
                    )
                )
                if not guild_data["shuffle"] and queue_dur > 0:
                    if query.is_local and query.is_album:
                        footer = _("folder")
                    else:
                        footer = _("search")

                    songembed.set_footer(
                        text=_(
                            "{time} until start of {type} playback: starts at #{position} in queue"
                        ).format(
                            time=queue_total_duration,
                            position=before_queue_length + 1,
                            type=footer,
                        )
                    )
                return await self.send_embed_msg(ctx, embed=songembed)
            elif query.is_local and query.single_track:
                tracks = await self.get_localtrack_folder_list(ctx, query)
            elif query.is_local and query.is_album:
                if ctx.invoked_with == "folder":
                    return await self._local_play_all(ctx, query, from_search=True)
                else:
                    tracks = await self.get_localtrack_folder_list(ctx, query)
            else:
                try:
                    result, called_api = await self.api_interface.fetch_track(ctx, player, query)
                except TrackEnqueueError:
                    self.update_player_lock(ctx, False)
                    return await self.send_embed_msg(
                        ctx,
                        title=_("Unable to Get Track"),
                        description=_(
                            "I'm unable to get a track from Lavalink at the moment, "
                            "try again in a few minutes."
                        ),
                    )
                except Exception as e:
                    self.update_player_lock(ctx, False)
                    raise e
                tracks = result.tracks
            if not tracks:
                embed = discord.Embed(title=_("Nothing found."))
                if await self.config.use_external_lavalink() and query.is_local:
                    embed.description = _(
                        "Local tracks will not work "
                        "if the `Lavalink.jar` cannot see the track.\n"
                        "This may be due to permissions or because Lavalink.jar is being run "
                        "in a different machine than the local tracks."
                    )
                elif query.is_local and query.suffix in _PARTIALLY_SUPPORTED_MUSIC_EXT:
                    embed = discord.Embed(title=_("Track is not playable."))
                    embed.description = _(
                        "**{suffix}** is not a fully supported format and some "
                        "tracks may not play."
                    ).format(suffix=query.suffix)
                return await self.send_embed_msg(ctx, embed=embed)
        else:
            tracks = query

        dj_enabled = self._dj_status_cache.setdefault(
            ctx.guild.id, await self.config.guild(ctx.guild).dj_enabled()
        )

        len_search_pages = math.ceil(len(tracks) / 5)
        search_page_list = []
        async for page_num in AsyncIter(range(1, len_search_pages + 1)):
            embed = await self._build_search_page(ctx, tracks, page_num)
            search_page_list.append(embed)

        if dj_enabled and not can_skip:
            return await menu(ctx, search_page_list)

        await menu(ctx, search_page_list, search_controls)
